# META
# {'passed': 2, 'failed': 0}
# END META

import os
import pytest

from collections import OrderedDict
from random import random

from pytest_harvest import saved_fixture
from pytest_harvest.common import yield_fixture

# init
this_file_name = os.path.split(__file__)[1]
unique_numbers = [random(), random()]


# The store is a global variable in this test
STORE = OrderedDict()


# Create our saved fixture, that will be saved in the global store
# -- 'normal' mode
@pytest.fixture(params=unique_numbers)
@saved_fixture(STORE)
def my_fix(request):
    # convert the parameter to a string so that the fixture is different from the parameter
    return str(request.param)


# -- 'generator' mode
@yield_fixture()
@saved_fixture(STORE)
def my_yield_fix():
    yield 12


def test_foo(my_fix, my_yield_fix):
    """"""
    print(my_fix)
    print(my_yield_fix)


# -----------------------------------------


def final_test(request):
    """This is the "test" that will be called when session ends. We check that the STORE contains everything"""
    for fixture_name, values in [('my_fix', [str(n) for n in unique_numbers]),
                                 ('my_yield_fix', [12, 12])]:
        assert fixture_name in STORE
        assert len(STORE[fixture_name]) == 2
        assert list(STORE[fixture_name].keys()) == [item.nodeid for item in request.session.items
                                                    if this_file_name in item.nodeid]
        assert list(STORE[fixture_name].values()) == values


@pytest.fixture(scope='session', autouse=True)
def register_final_test(request):
    # This is a way, compliant with legacy pytest 2.8, to register our teardown callback
    request.addfinalizer(lambda: final_test(request))



from collections import OrderedDict
from inspect import isgeneratorfunction

from pytest_harvest.decorator_hack import my_decorate


try:  # python 3+
    from typing import Union, Any, Dict
except ImportError:
    pass


try:
    from pytest_steps.steps_generator import get_underlying_fixture
except ImportError:
    def get_underlying_fixture(x):
        return x


def saved_fixture(store,    # type: Union[str, Dict[str, Any]]
                  key=None  # type: str
                  ):
    """
    Decorates a fixture so that it is saved in `store`. `store` can be a dict-like variable or a string
    representing a fixture name to a dict-like session-scoped fixture (TODO)

    After executing all tests, <store> will contain a new item under key <name>. This item is a dictionary
    <test_id>: <fixture_value> for each test node.

    ```python
    from random import random

    @saved_fixture('storage')
    @pytest.fixture
    def dummy():
         return random()
    ```

    :param store: a dict-like object or a fixture name corresponding to a dict-like object. in this dictionary, a new
        entry will be added for the fixture. This entry will contain a dictionary <test_id>: <fixture_value> for each
        test node.
    :param key: the name associated with the stored fixture in the store. By default this is the fixture name.
    :return: a fixture that will be stored
    """
    # trick to support both with-args and without args usage
    return _saved_fixture(store, key)


def _saved_fixture(*args, **kwargs):
    """ Inner decorator creation method to support no-arg calls """
    if len(args) == 1 and callable(args[0]):
        # called without arguments, directly decorates a function
        # Note: since store is currently mandatory, this never happens.
        # But we keep it in case of signature change
        f = args[0]
        return make_saved_fixture(f)
    else:
        # return a function decorator
        def _decorator(f):
            return make_saved_fixture(f, *args, **kwargs)
        return _decorator


# def get_fixture_name(fixture_fun):
#     """
#     Internal utility to retrieve the fixture name corresponding to the given fixture function .
#     Indeed there is currently no pytest API to do this.
#
#     :param fixture_fun:
#     :return:
#     """
#     custom_fixture_name = getattr(fixture_fun._pytestfixturefunction, 'name', None)
#
#     if custom_fixture_name is not None:
#         # there is a custom fixture name
#         return custom_fixture_name
#     else:
#         obj__name = getattr(fixture_fun, '__name__', None)
#         if obj__name is not None:
#             # a function, probably
#             return obj__name
#         else:
#             # a callable object probably
#             return str(fixture_fun)


def make_saved_fixture(fixture_fun,
                       store,    # type: Union[str, Dict[str, Any]]
                       key=None  # type: str
                       ):
    """
    Manual decorator to decorate a (future) fixture function so that it is saved in a storage.

    ```python
    def f():
        return "dummy"

    # manually declare that it should be saved
    f = make_saved_fixture(f)

    # manually make a fixture from f
    f = pytest.fixture(f)
    ```

    See `@saved_fixture` decorator for parameter details.
    """

    # Simply get the function name
    fixture_name = getattr(fixture_fun, '__name__', None) or str(fixture_fun)

    # Name to use for storage
    key = key or fixture_name

    # is the store a fixture or an object ?
    store_is_a_fixture = isinstance(store, str)

    # if the store object is already available, we can ensure that it is initialized. Otherwise trust pytest for that
    if not store_is_a_fixture:
        if key in store.keys():
            raise ValueError("Key '%s' already exists in store object. Please make sure that your store object is "
                             "properly initialized as an empty dict-like object, and/or provide a different custom "
                             "`name` if two stored fixtures share the same storage key name.")

    # Note: we can not init the storage[key] entry here because when storage is a fixture, it does not yet exist.

    def _init_and_check(request, store):
        # Init storage if needed
        if key not in store:
            store[key] = OrderedDict()
        # Check that the node id is unique
        if request.node.nodeid in store[key]:
            raise KeyError("Internal Error - This fixture '%s' was already "
                           "stored for test id '%s'" % (key, request.node.nodeid))

    # Wrap the fixture in the correct mode (generator or not)
    if not isgeneratorfunction(fixture_fun):
        def fixture_wrapper(f, request, *args, **kwargs):
            """Wraps a fixture so as to store it before it is returned"""
            # get the actual store object
            if store_is_a_fixture:
                store_ = args[0]
                args = args[1:]
            else:
                # use the variable from outer scope (from `make_saved_fixture`)
                store_ = store
            _init_and_check(request, store_)
            fixture_value = f(*args, **kwargs)                                       # Get the fixture
            store_[key][request.node.nodeid] = get_underlying_fixture(fixture_value)  # Store it
            return fixture_value                                                     # Return it

    else:
        def fixture_wrapper(f, request, *args, **kwargs):
            """Wraps a fixture so as to store it before it is returned (generator mode)"""
            # get the actual store object
            if store_is_a_fixture:
                store_ = args[0]
                args = args[1:]
            else:
                # use the variable from outer scope (from `make_saved_fixture`)
                store_ = store
            _init_and_check(request, store_)
            gen = f(*args, **kwargs)
            fixture_value = next(gen)                                                # Get the fixture
            store_[key][request.node.nodeid] = get_underlying_fixture(fixture_value)  # Store it
            yield fixture_value                                                      # Return it

            # Make sure to terminate the underlying generator
            try:
                next(gen)
            except StopIteration:
                pass

    if store_is_a_fixture:
        # add 'store' as second positional argument
        stored_fixture_function = my_decorate(fixture_fun, fixture_wrapper, additional_args=('request', store))
    else:
        stored_fixture_function = my_decorate(fixture_fun, fixture_wrapper, additional_args=('request', ))

    return stored_fixture_function



from collections import OrderedDict
from itertools import chain
from warnings import warn

try: # python 3.5+
    from typing import Union, Iterable, Mapping, Any
except ImportError:
    pass

from pytest_harvest.plugin import HARVEST_PREFIX


PYTEST_OBJ_NAME = 'pytest_obj'


def get_session_synthesis_dct(session,
                              status_details=False,    # type: bool
                              durations_in_ms=False,   # type: bool
                              pytest_prefix=None,      # type: bool
                              filter=None,             # type: Any
                              filter_incomplete=True,  # type: bool
                              flatten=False,           # type: bool
                              fixture_store=None,      # type: Union[Mapping[str, Any], Iterable[Mapping[str, Any]]]
                              flatten_more=None        # type: Union[str, Iterable[str], Mapping[str, str]]
                              ):
    """
    Returns a dictionary containing a synthesis of what is available currently in the `pytest` session object provided.
     - 'pytest_obj': the object under test, typically a test function
     - 'pytest_status': the overall status ('failing', 'skipped', 'passed')
     - 'pytest_duration': the duration of the 'call' step. By default this is the pytest unit (s) but if you set
     `durations_in_ms=True` it becomes (ms)
     - 'pytest_status_details': a dictionary containing step-by-step status details for all pytest steps ('setup',
     'call', 'teardown'). This is only included if `status_details=True` (not by default)

    The 'pytest' prefix in front of all these items (except `pytest_obj`) is by default added in non-flatten mode and
    removed in flatten mode. To force one of these you can set `pytest_prefix` to True or False.

    An optional `filter` can be provided, that can be a singleton or iterable of pytest objects (typically test
    functions). It can also be a module, or the special `THIS_MODULE` object.

    If this method is called before the end of the pytest session, some nodes might be incomplete, i.e. they will not
    have data for the three stages (setup/call/teardown). By default these nodes are filtered out but you can set
    `filter_incomplete=False` to make them appear. They will have a special 'pending' synthesis status.

    An optional collection of storage objects can be provided, so as to merge them into the resulting dictionary.

    Finally a `flatten_output` option allows users to get a flat dictionary output instead of nested status details,
    parameters dict, and storage dicts.

    :param session: a pytest session object.
    :param status_details: a flag indicating if pytest status details per stage (setup/call/teardown) should be
        included. Default=`False`: only the pytest status summary is provided.
    :param durations_in_ms: by default `pytest` measures durations in seconds so they are outputed in this unit. You
        can turn the flag to True to output milliseconds instead.
    :param pytest_prefix: to add (True) or remove (False) the 'pytest_' prefix in front of status, duration and status
        details. Typically useful in flatten mode when the names are not ambiguous. By default it is None, which
        means =(not flatten)
    :param filter: a singleton or iterable of pytest objects on which to filter the returned dict on (the returned
        items will only by pytest nodes for which the pytest object is one of the ones provided). One can also use
        modules or the special `THIS MODULE` item.
    :param filter_incomplete: a boolean indicating if incomplete nodes (without the three stages setup/call/teardown)
        should appear in the results (False) or not (True, default).
    :param flatten: a boolean (default `False`) indicating if the resulting dictionary should be flattened. If it
        set to `True`, the 3 nested dictionaries (pytest status details, parameters, and optionally storages)
        will have their contents directly copied in the first level (with a prefix added in case of pytest status
        details).
    :param fixture_store: a singleton or iterable containing dict-like fixture storage objects (see
        `@saved_fixture` and `create_results_bag`). If flatten=`False` the contents of these dictionaries will be added
        to the output in a dedicated 'fixtures' entry. If flatten=True all of their contents will be included directly.
    :param flatten_more: a singleton, iterable or dictionary containing fixture names to flatten one level more in case
        flatten=True. If a dictionary is provided, the key should be the fixture name, and the value should be a prefix
        used for flattening its contents
    :return: a dictionary where the keys are pytest node ids. Each value is also a dictionary, containing information
        available from pytest concerning the test node, and optionally storage contents if `storage_dcts` is provided.
    """
    res_dct = OrderedDict()

    # Optional 'pytest_' prefix in front of status and duration
    if pytest_prefix is None:
        pytest_prefix = not flatten
    if pytest_prefix:
        pytest_prefix = 'pytest_'
    else:
        pytest_prefix = ''

    # Optional filter
    if filter is not None:
        if isinstance(filter, str):
            filter = {filter}
        else:
            try:
                iter(filter)
                filter = set(filter)
            except TypeError:
                # TypeError: '<....>' object is not iterable
                filter = {filter}
        def is_selected(item_obj):
            if item_obj in filter:
                return True
            elif any(item_obj.__module__ == f for f in filter):
                return True
            else:
                return False
        filtered_items = (item for item in session.items if is_selected(item.obj))
    else:
        filtered_items = session.items

    # fixture store check
    if fixture_store is not None:
        try:
            fixture_store_items = fixture_store.items()
        except AttributeError:
            # not a dict: an iterable of dict
            fixture_store_items = list(chain(store.items() for store in fixture_store))

    # flatten_more check
    if flatten_more is not None:
        if isinstance(flatten_more, dict):
            flatten_more_prefixes_dct = flatten_more.items()
        elif isinstance(flatten_more, str):
            # single name ?
            flatten_more_prefixes_dct = {flatten_more: ''}
        else:
            # iterable ?
            flatten_more_prefixes_dct = {k: '' for k in flatten_more}

    # For each item add an entry
    for item in filtered_items:
        item_dct = OrderedDict()

        # Fill the dictionary with information about this test node
        # -- test object
        item_dct[PYTEST_OBJ_NAME] = item.obj

        # -- test status: this information is available thanks to our hook in plugin.py
        (test_status, test_duration), status_dct = get_pytest_status(item, durations_in_ms=durations_in_ms)

        if test_status != 'pending' or not filter_incomplete:
            # -- parameters (of tests and fixtures)
            param_dct = get_pytest_params(item)

            # Fill according to mode
            item_dct[pytest_prefix + "status"] = test_status
            item_dct[pytest_prefix + "duration_" + ('ms' if durations_in_ms else 's')] = test_duration
            if flatten:
                if status_details:
                    for k, v in status_dct.items():
                        item_dct[pytest_prefix + "status__" + k] = v
                item_dct.update(param_dct)
            else:
                if status_details:
                    item_dct[pytest_prefix + "status_details"] = status_dct
                item_dct[pytest_prefix + "params"] = param_dct

            # -- fixture storages
            # For info: https://docs.pytest.org/en/latest/_modules/_pytest/runner.html
            # used_fixtures = sorted(item._fixtureinfo.name2fixturedefs.keys())
            # if used_fixtures:
            #     tw.write(" (fixtures used: {})".format(", ".join(used_fixtures)))
            if fixture_store is not None:
                if not flatten:
                    item_dct['fixtures'] = OrderedDict()

                for fixture_name, fixture_dct in fixture_store_items:

                    # if this fixture is available for this test
                    if item.nodeid in fixture_dct:
                        # get the fixture value for this test
                        fix_val = fixture_dct[item.nodeid]

                        # store it in the appropriate format
                        if flatten:
                            if flatten_more is not None and fixture_name in flatten_more_prefixes_dct:
                                prefix = flatten_more_prefixes_dct[fixture_name]
                                # flatten more
                                for k, v in fix_val.items():
                                    item_dct[prefix + k] = v
                            else:
                                item_dct[fixture_name] = fix_val
                        else:
                            item_dct['fixtures'][fixture_name] = fix_val

            # Finally store in the main dictionary
            res_dct[item.nodeid] = item_dct

    return res_dct


def get_pytest_status(item, durations_in_ms):
    """ Returns a dictionary containing item's pytest status (success/skipped/failed, duration converted to ms) """

    # the status keys that have been stored by our plugin.py module
    status_keys = [k for k in vars(item) if k.startswith(HARVEST_PREFIX)]
    if len(status_keys) == 0:
        warn("[pytest-harvest] Test items status is not available. You should maybe install pytest-harvest with "
             "pip. If it is already the case, you case try to force-use it by adding "
             "`pytest_plugins = ['harvest']` to your conftest.py. But for normal use this should not be required, "
             "installing with pip should be enough.")

    # adjust duration factor according to target unit
    duration_factor = (1000 if durations_in_ms else 1)

    # create the status dictionary for that item
    status_dct = OrderedDict()
    test_status = 'passed'
    test_duration = None
    for k in status_keys:
        statusreport = getattr(item, k)
        status_dct[statusreport.when] = (statusreport.outcome, statusreport.duration * duration_factor)
        # update global test status
        if test_status == 'passed' \
                or (test_status == 'skipped' and statusreport.outcome != 'passed'):
            test_status = statusreport.outcome
        # global test duration is the duration of the "call" step only
        if statusreport.when == "call":
            test_duration = statusreport.duration * duration_factor

    if len(status_keys) < 3:
        # this is an incomplete test
        test_status = 'pending'

    return (test_status, test_duration), status_dct


def get_pytest_params(item):
    """ Returns a dictionary containing item's parameters """

    param_dct = OrderedDict()
    for param_name in item.fixturenames:
        if hasattr(item, 'callspec'):
            if param_name in item.callspec.params:
                # this is a param: ok
                # IMPORTANT: fixture parameters have the same name than the fixtures themselves!
                param_dct[param_name] = item.callspec.params[param_name]
            else:
                # this is a fixture: it is not available by default in item, this is normal pytest behaviour
                # (hence the @saved_fixture decorator)
                pass

    return param_dct



import ast
import os
import re
from os.path import join, dirname, pardir

import pytest
import six

# Make the list of all tests that we will have to execute (each in an independent pytest runner)
THIS_DIR = dirname(__file__)
tests_raw_folder = join(THIS_DIR, pardir, 'tests_raw')
test_files = [f for f in os.listdir(tests_raw_folder) if f.startswith('test')]


META_REGEX = re.compile(
"""^(# META
# )(?P<asserts_dct>.*)
(# END META)
.*""")


@pytest.mark.parametrize('test_to_run', test_files, ids=str)
def test_run_all_tests(test_to_run, testdir):
    """
    This is a meta-test. It is executed for each test file in the 'tests_raw' folder.
    For each of them, the file is retrieved and the expected test results are read from its first lines.
    Then a dedicated pytest runner is run on this file, and the results are compared with the expected ones.

    See https://docs.pytest.org/en/latest/writing_plugins.html

    :param test_to_run:
    :param testdir:
    :return:
    """

    with open(join(tests_raw_folder, test_to_run)) as f:
        # Create a temporary conftest.py file
        # testdir.makeconftest("""""")

        # create a temporary pytest test file
        test_file_contents = f.read()
        testdir.makepyfile(test_file_contents)

        # Grab the expected things to check when this is executed
        m = META_REGEX.match(test_file_contents)
        assert m is not None
        asserts_dct_str = m.groupdict()['asserts_dct']
        asserts_dct = ast.literal_eval(asserts_dct_str)

        # Here we run pytest
        print("\nTesting that running pytest on file %s results in %s" % (test_to_run, str(asserts_dct)))
        result = testdir.runpytest()  # ("-q")

        # Here we check that everything is ok
        try:
            result.assert_outcomes(**asserts_dct)
        except Exception as e:
            print("Error while asserting that %s results in %s" % (test_to_run, str(asserts_dct)))
            six.raise_from(e, e)



import sys
from itertools import chain

from decorator import FunctionMaker
from inspect import isgeneratorfunction

try:  # python 3.3+
    from inspect import signature
except ImportError:
    from funcsigs import signature

try:
    from decorator import iscoroutinefunction
except ImportError:
    try:
        from inspect import iscoroutinefunction
    except ImportError:
        # let's assume there are no coroutine functions in old Python
        def iscoroutinefunction(f):
            return False


class MyFunctionMaker(FunctionMaker):
    """
    Overrides FunctionMaker so that additional arguments can be inserted in the resulting signature.
    """

    def refresh_signature(self):
        """Update self.signature and self.shortsignature based on self.args,
        self.varargs, self.varkw"""
        allargs = list(self.args)
        allshortargs = list(self.args)
        if self.varargs:
            allargs.append('*' + self.varargs)
            allshortargs.append('*' + self.varargs)
        elif self.kwonlyargs:
            allargs.append('*')  # single star syntax
        for a in self.kwonlyargs:
            allargs.append('%s=None' % a)
            allshortargs.append('%s=%s' % (a, a))
        if self.varkw:
            allargs.append('**' + self.varkw)
            allshortargs.append('**' + self.varkw)
        self.signature = ', '.join(allargs)
        self.shortsignature = ', '.join(allshortargs)

    @classmethod
    def create(cls, obj, body, evaldict, defaults=None,
               doc=None, module=None, addsource=True, add_args=(), **attrs):
        """
        Create a function from the strings name, signature and body.
        evaldict is the evaluation dictionary. If addsource is true an
        attribute __source__ is added to the result. The attributes attrs
        are added, if any.

        If add_args is not empty, these arguments will be prepended to the
        positional arguments.
        """
        if isinstance(obj, str):  # "name(signature)"
            name, rest = obj.strip().split('(', 1)
            signature = rest[:-1]  # strip a right parens
            func = None
        else:  # a function
            name = None
            signature = None
            func = obj
        self = cls(func, name, signature, defaults, doc, module)
        ibody = '\n'.join('    ' + line for line in body.splitlines())
        caller = evaldict.get('_call_')  # when called from `decorate`
        if caller and iscoroutinefunction(caller):
            body = ('async def %(name)s(%(signature)s):\n' + ibody).replace(
                'return', 'return await')
        else:
            body = 'def %(name)s(%(signature)s):\n' + ibody

            # Handle possible signature changes
            sig_modded = False
            if len(add_args) > 0:
                # prepend them as positional args - hence the reversed()
                for arg in reversed(add_args):
                    if arg not in self.args:
                        self.args = [arg] + self.args
                        sig_modded = True
                    else:
                        # the argument already exists in the wrapped
                        # function, nothing to do.
                        pass
            if sig_modded:
                self.refresh_signature()

            # make the function
            func = self.make(body, evaldict, addsource, **attrs)

            if sig_modded:
                # delete this annotation otherwise inspect.signature
                # will wrongly return the signature of func.__wrapped__
                # instead of the signature of func
                del func.__wrapped__

            return func


def _extract_additional_args(f_sig, add_args_names, args, kwargs):
    """
    Processes the arguments received by our caller so that at the end, args
    and kwargs only contain what is needed by f (according to f_sig). All
    additional arguments are returned separately, in order described by
    `add_args_names`. If some names in `add_args_names` are present in `f_sig`,
    then the arguments will appear both in the additional arguments and in
    *args, **kwargs.

    In the end, only *args can possibly be modified by the procedure (by removing
    from it all additional arguments that were not in f_sig and were prepended).

    So the result is a tuple (add_args, args)

    :return: a tuple (add_args, args) where `add_args` are the values of
        arguments named in `add_args_names` in the same order ; and `args` is
        the positional arguments to send to the wrapped function together with
        kwargs (args now only contains the positional args that are required by
        f, without the extra ones)
    """
    # -- first the 'truly' additional ones (the ones not in the signature)
    add_args = [None] * len(add_args_names)
    for i, arg_name in enumerate(add_args_names):
        if arg_name not in f_sig.parameters:
            # remove this argument from the args and put it in the right place
            add_args[i] = args[0]
            args = args[1:]

    # -- then the ones that already exist in the signature. Thanks,inspect pkg!
    bound = f_sig.bind(*args, **kwargs)
    for i, arg_name in enumerate(add_args_names):
        if arg_name in f_sig.parameters:
            add_args[i] = bound.arguments[arg_name]

    return add_args, args


def _wrap_caller_for_additional_args(func, caller, additional_args):
    """
    This internal function wraps the caller so as to handle all cases
    (if some additional args are already present in the signature or not)
    so as to ensure a consistent caller signature.

    :return: a new caller wrapping the caller, to be used in `decorate`
    """
    f_sig = signature(func)

    # We will create a caller above the original caller in order to check
    # if additional_args are already present in the signature or not, and
    # act accordingly
    original_caller = caller

    # -- then create the appropriate function signature according to
    # wrapped function signature assume that original_caller has all
    # additional args as first positional arguments, in order
    if not isgeneratorfunction(original_caller):
        def caller(f, *args, **kwargs):
            # Retrieve the values for additional args.
            add_args, args = _extract_additional_args(f_sig, additional_args,
                                                      args, kwargs)

            # Call the original caller
            return original_caller(f, *chain(add_args, args),
                                   **kwargs)
    else:
        def caller(f, *args, **kwargs):
            # Retrieve the value for additional args.
            add_args, args = _extract_additional_args(f_sig, additional_args,
                                                      args, kwargs)

            # Call the original caller
            for res in original_caller(f, *chain(add_args, args),
                                       **kwargs):
                yield res

    return caller


def my_decorate(func, caller, extras=(), additional_args=()):
    """
    decorate(func, caller) decorates a function using a caller.
    If the caller is a generator function, the resulting function
    will be a generator function.

    You can provide additional arguments with `additional_args`. In that case
    the caller's signature should be

        `caller(f, <additional_args_in_order>, *args, **kwargs)`.

    `*args, **kwargs` will always contain the arguments required by the inner
    function `f`. If `additional_args` contains argument names that are already
    present in `func`, they will be present both in <additional_args_in_order>
    AND in `*args, **kwargs` so that it remains easy for the `caller` both to
    get the additional arguments' values directly, and to call `f` with the
    right arguments.
    """
    if len(additional_args) > 0:
        # wrap the caller so as to handle all cases
        # (if some additional args are already present in the signature or not)
        # so as to ensure a consistent caller signature
        caller = _wrap_caller_for_additional_args(func, caller, additional_args)

    evaldict = dict(_call_=caller, _func_=func)
    es = ''
    for i, extra in enumerate(extras):
        ex = '_e%d_' % i
        evaldict[ex] = extra
        es += ex + ', '

    if '3.5' <= sys.version < '3.6':
        # with Python 3.5 isgeneratorfunction returns True for all coroutines
        # however we know that it is NOT possible to have a generator
        # coroutine in python 3.5: PEP525 was not there yet
        generatorcaller = isgeneratorfunction(
            caller) and not iscoroutinefunction(caller)
    else:
        generatorcaller = isgeneratorfunction(caller)
    if generatorcaller:
        fun = MyFunctionMaker.create(
            func, "for res in _call_(_func_, %s%%(shortsignature)s):\n"
                  "    yield res" % es, evaldict,
            add_args=additional_args, __wrapped__=func)
    else:
        fun = MyFunctionMaker.create(
            func, "return _call_(_func_, %s%%(shortsignature)s)" % es,
            evaldict, add_args=additional_args, __wrapped__=func)
    if hasattr(func, '__qualname__'):
        fun.__qualname__ = func.__qualname__
    return fun



import pytest


HARVEST_PREFIX = "harvest_rep_"


@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_runtest_makereport(item, call):
    """
    We use this hook to store the test execution status for each node in the 'item' object,
    so that we can use it in `get_session_synthesis_dct`

    It is called because the whole package is a pytest plugin (it has a pytest entry point in setup.py)

    Following the example here:
    https://docs.pytest.org/en/latest/example/simple.html#making-test-result-information-available-in-fixtures

    :param item:
    :param call:
    :return:
    """
    # execute all other hooks to obtain the report object
    outcome = yield
    rep = outcome.get_result()

    # set a report attribute for each phase of a call, which can
    # be "setup", "call", "teardown"

    setattr(item, HARVEST_PREFIX + rep.when, rep)



# META
# {'passed': 8, 'skipped': 0, 'failed': 0}
# END META
from collections import OrderedDict
from random import random

import pandas as pd
import pytest
from tabulate import tabulate

from pytest_harvest import create_results_bag_fixture, saved_fixture, get_session_synthesis_dct


# ---------- The function to test -------
def my_algorithm(param, data):
    # let's return a random accuracy !
    return random()


# ---------- Tests
# A module-scoped store
@pytest.fixture(scope='module', autouse=True)
def store():
    return OrderedDict()


# A module-scoped results bag fixture
my_results = create_results_bag_fixture('store', name='my_results')


@pytest.fixture(params=['A', 'B', 'C'])
@saved_fixture('store')
def dataset(request):
    """Represents a dataset fixture."""
    return "my dataset #%s" % request.param


@pytest.mark.parametrize("algo_param", [1, 2], ids=str)
def test_my_app_bench(algo_param, dataset, my_results):
    """
    This test applies the algorithm with various parameters (`algo_param`)
    on various datasets (`dataset`).

    Accuracies are stored in a results bag (`results_bag`)
    """
    # apply the algorithm with param `algo_param` on dataset `dataset`
    accuracy = my_algorithm(algo_param, dataset)
    # store it in the results bag
    my_results.accuracy = accuracy


def test_basic():
    """Another test, to show how it appears in the results"""
    pass


# Our test synthesis
def test_synthesis(request, store):
    """
    An example test that retrieves synthesis information about this module
    """
    # retrieve the synthesis, merged with the fixture store
    results_dct = get_session_synthesis_dct(request.session, filter=test_synthesis.__module__,
                                            durations_in_ms=True,
                                            status_details=False, fixture_store=store,
                                            flatten=True, flatten_more='my_results')

    # print keys and first node details
    print("\nKeys:\n" + "\n".join(list(results_dct.keys())))
    print("\nFirst node:\n" + "\n".join(repr(k) + ": " + repr(v) for k, v in list(results_dct.values())[0].items()))

    # convert to a pandas dataframe
    results_df = pd.DataFrame.from_dict(results_dct, orient='index')
    results_df.index = results_df.index.to_series().apply(lambda test_id: test_id.split('::')[-1])  # remove full path
    results_df.drop(['pytest_obj'], axis=1, inplace=True)                                  # drop pytest object column

    # print using tabulate
    print(tabulate(results_df, headers='keys'))


# ------- Output -------
#
#                         status      duration_ms    algo_param  dataset           accuracy
# ----------------------  --------  -------------  ------------  -------------  -----------
# test_basic              passed         0.999928           nan  nan            nan
# test_my_app_bench[A-1]  passed         0                    1  my dataset #A    0.818458
# test_my_app_bench[A-2]  passed         0                    2  my dataset #A    0.0364919
# test_my_app_bench[B-1]  passed         0                    1  my dataset #B    0.0885096
# test_my_app_bench[B-2]  passed         1.0004               2  my dataset #B    0.826001
# test_my_app_bench[C-1]  passed         0                    1  my dataset #C    0.700515
# test_my_app_bench[C-2]  passed         0                    2  my dataset #C    0.281405
#



import requests
import shutil
from os import makedirs, path
import xunitparser


def download_badge(junit_xml: str='reports/junit/junit.xml', dest_folder: str='reports/junit'):

    makedirs(dest_folder, exist_ok=True)

    # read the junit test file
    ts, tr = xunitparser.parse(open(junit_xml))
    runned = tr.testsRun
    failed = len(tr.failures)

    success_percentage = round((runned - failed) * 100 / runned)
    if success_percentage < 50:
        color = 'red'
    elif success_percentage < 75:
        color = 'orange'
    elif success_percentage < 90:
        color = 'green'
    else:
        color = 'brightgreen'
    url = 'https://img.shields.io/badge/tests-' + str(success_percentage) + '%25-' + color + '.svg'

    dest_file = path.join(dest_folder, 'junit-badge.svg')

    print('Generating junit badge from : ' + url)
    response = requests.get(url, stream=True)
    with open(dest_file, 'wb') as out_file:
        response.raw.decode_content = True
        shutil.copyfileobj(response.raw, out_file)
    del response


if __name__ == "__main__":
    # execute only if run as a script
    download_badge()


# META
# {'passed': 6, 'skipped': 0, 'failed': 0}
# END META
from collections import OrderedDict
from random import random
import pytest

from pytest_harvest import create_results_bag_fixture, get_session_synthesis_dct
from pytest_harvest.common import yield_fixture


def my_algorithm(param, data):
    # let's return a random accuracy !
    return random()


@pytest.fixture(params=['A', 'B', 'C'])
def dataset(request):
    return "my dataset #%s" % request.param


@pytest.mark.parametrize("algo_param", [1, 2], ids=str)
def test_my_app_bench(algo_param, dataset, results_bag):
    """
    This test applies the algorithm with various parameters (`algo_param`)
    on various datasets (`dataset`). Accuracies are stored in a results
    bag (`results_bag`)
    """
    # apply the algorithm with param `algo_param` on dataset `dataset`
    accuracy = my_algorithm(algo_param, dataset)

    # store it in the results bag
    results_bag.accuracy = accuracy


# the result bag
# note: depending on your pytest version, the name used by pytest might be
# the variable name (left) or the one you provide in the 'name' argument so
# make sure they are identical!
results_bag = create_results_bag_fixture('store', name="results_bag")


@yield_fixture(scope='module', autouse=True)
def store(request):
    # setup: init the store
    store = OrderedDict()
    yield store
    # teardown: here you can collect all
    assert len(store['results_bag']) == 6
    print(dict(store['results_bag']))

    # retrieve the synthesis, merged with the fixture store
    results_dct = get_session_synthesis_dct(request.session, status_details=False, durations_in_ms=True,
                                            fixture_store=store, flatten=True, flatten_more='results_bag')

    # -- use pandas to print
    import pandas as pd
    results_df = pd.DataFrame.from_dict(results_dct, orient='index')
    # (a) remove the full test id path
    results_df.index = results_df.index.to_series().apply(lambda test_id: test_id.split('::')[-1])
    # (b) drop pytest object column
    results_df.drop(['pytest_obj'], axis=1, inplace=True)

    from tabulate import tabulate
    print(tabulate(results_df, headers='keys', tablefmt="pipe").replace(':-', '--').replace('-:', '--'))



"""
equivalent of pip install -r <file>
 - with environment variables replacement
 - and all dependencies are installed in one 'pip install' call (solving potential complex deps)
"""
import os

import re
import sys
import subprocess


def check_cmd(cmd):
    assert isinstance(cmd, str), "cmd should be a string"
    assert cmd in {"pip", "conda"}, "cmd should be conda or pip. Unknown: " + str(cmd)


def install(cmd, packages):
    """
    Installs all packages provided at once
    :param packages:
    :return:
    """
    check_cmd(cmd)

    all_pkgs_str = " ".join(all_pkgs)
    print("INSTALLING: " + cmd + " install " + all_pkgs_str)
    subprocess.check_call([cmd, 'install'] + packages) # install pkg


env_var_regexp = re.compile(".*\$(\S+).*")


if __name__ == '__main__':
    assert len(sys.argv[1:]) >= 2, "at least two mandatory arguments: <cmd> <filename>"

    cmd = sys.argv[1]
    check_cmd(cmd)

    filenames = sys.argv[2:]

    all_pkgs = []
    for filename in filenames:
        with open(filename) as f:
            for line in f.readlines():
                # First remove any comment on that line
                splitted = line.split('#', 1)  # (maxsplit=1) but python 2 does not support it :)
                splitted = splitted[0].strip().rstrip()
                if splitted != '':
                    # the replace env vars
                    env_var_found=True
                    while env_var_found:
                        res = env_var_regexp.match(splitted)
                        env_var_found = res is not None
                        if env_var_found:
                            env_var_name = res.groups()[0]
                            try:
                                env_var_val = os.environ[env_var_name]
                                print("replacing $%s with %s" % (env_var_name, env_var_val))
                                splitted = splitted.replace("$%s" % env_var_name, env_var_val)
                            except KeyError:
                                raise Exception("Environment variable does not exist in file %s: $%s"
                                                "" % (filename, env_var_name))
                        else:
                            all_pkgs.append(splitted)

    install(cmd, all_pkgs)



# META
# {'passed': 9, 'skipped': 0, 'failed': 0}
# END META
from collections import OrderedDict
from itertools import product

import pytest

from pytest_harvest import create_results_bag_fixture, saved_fixture, get_session_synthesis_dct


@pytest.fixture(scope='module', autouse=True)
def store():
    return OrderedDict()


my_results = create_results_bag_fixture('store', name='my_results')


# parametrized fixture
fixture_params = ['A', 'B', 'C']


@pytest.fixture(params=fixture_params)
@saved_fixture('store')
def my_fix(request):
    """Our saved fixture, that will be saved in the store fixture"""
    # convert the parameter to a string so that the fixture is different from the parameter
    return "my_fix #%s" % request.param


# parametrized test
test_params = [10, 20]


@pytest.mark.parametrize('p', test_params, ids="p={}".format)
def test_complete(p, my_fix, my_results):
    """This fake test has a parameter, a fixture and stores things in a results bag"""
    my_results.score = p * 10
    my_results.what = 'hello ' + my_fix


@pytest.mark.parametrize('flatten, flatten_more', [(False, None), (True, None), (True, 'my_results')],
                         ids="flatten={}, flatten_more={}".format)
def test_synthesis(flatten, flatten_more, request, store):
    """Tests that the synthesis dictionary combined with with fixture store is ok"""

    # retrieve the synthesis, merged with the fixture store
    results_dct = get_session_synthesis_dct(request.session, filter=test_complete, status_details=False,
                                            fixture_store=store, flatten=flatten, flatten_more=flatten_more)

    # ------ PRINTS ---------

    # --test node ids
    print("\n".join(list(results_dct.keys())))

    # --zoom on first node
    print("\n".join(repr(k) + ": " + repr(v) for k, v in list(results_dct.values())[0].items()))

    if flatten and flatten_more:
        # -- use tabulate to print
        from tabulate import tabulate
        # print(tabulate(dct, headers='keys'))  not possible, it does not yet support that dict keys represent rows

        # -- use pandas to print
        import pandas as pd
        results_df = pd.DataFrame.from_dict(results_dct, orient='index')
        # (a) remove the full test id path
        results_df.index = results_df.index.to_series().apply(lambda test_id: test_id.split('::')[-1])
        # (b) replace pytest object with its name
        results_df['pytest_obj'] = results_df['pytest_obj'].map(lambda f: f.__name__)

        print(tabulate(results_df, headers='keys', tablefmt="pipe").replace(':-', '--').replace('-:', '--'))

    # ------ ASSERTS ---------

    # compute the parameter values for all tests in order
    params = list(product(fixture_params, test_params))
    assert len(results_dct) == len(params)

    for i, (nodeid, node_info) in enumerate(results_dct.items()):
        if flatten:
            where_dct = node_info
            if flatten_more is None:
                where_dct_results = node_info['my_results']
            else:
                where_dct_results = node_info

        else:
            where_dct = node_info['fixtures']
            where_dct_results = node_info['fixtures']['my_results']

        assert where_dct['my_fix'] == 'my_fix #' + params[i][0]
        assert where_dct_results['score'] == params[i][1] * 10
        assert where_dct_results['what'] == 'hello my_fix #' + params[i][0]



from pytest_harvest.common import get_fixture_value
from pytest_harvest.fixture_cache import saved_fixture, make_saved_fixture
from pytest_harvest.results_bags import create_results_bag_fixture
from pytest_harvest.results_session import get_session_synthesis_dct, PYTEST_OBJ_NAME

__all__ = [
    # submodules
    'fixture_cache', 'results_bags', 'results_session',

    # symbols imported above
    'get_fixture_value',
    'saved_fixture', 'make_saved_fixture',
    'create_results_bag_fixture',
    'get_session_synthesis_dct', 'PYTEST_OBJ_NAME',
    ]



import pytest


def get_fixture_value(request, fixture_name):
    """
    Returns the value associated with fixture named `fixture_name`, in provided request context.
    This is just an easy way to use `getfixturevalue` or `getfuncargvalue` according to whichever is availabl in
    current pytest version

    :param request:
    :param fixture_name:
    :return:
    """
    try:
        # Pytest 4+ or latest 3.x (to avoid the deprecated warning)
        return request.getfixturevalue(fixture_name)
    except AttributeError:
        # Pytest 3-
        return request.getfuncargvalue(fixture_name)


# Create a symbol that will work to create a fixture containing 'yield', whatever the pytest version
if int(pytest.__version__.split('.', 1)[0]) < 3:
    yield_fixture = pytest.yield_fixture
else:
    yield_fixture = pytest.fixture



from datetime import datetime

import pytest
from six import raise_from

try:  # python 3+
    from typing import Type, Set, Union, Any, Dict
except ImportError:
    pass

try:
    from pytest_harvest import one_per_step
except ImportError:
    # pytest steps not installed: define as a transparent decorator
    def one_per_step(f):
        return f

from pytest_harvest.common import yield_fixture
from pytest_harvest.fixture_cache import saved_fixture


class ResultsBag(dict):
    """
    A 'Munch', that is, a dual object/dict.
    """

    # object
    def __setattr__(self, key, value):
        if key != '__dct__':
            try:
                self[key] = value
            except KeyError as e:
                raise_from(AttributeError(key), e)
        else:
            object.__setattr__(self, key, value)

    def __getattr__(self, key):
        if key != '__dct__':
            try:
                return self[key]
            except KeyError as e:
                raise_from(AttributeError(key), e)
        else:
            return object.__getattribute__(self, key)

    def __delattr__(self, key):
        if key != '__dct__':
            try:
                del self[key]
            except KeyError as e:
                raise_from(AttributeError(key), e)
        else:
            raise ValueError("Can not delete __dct__: read-only")

    # # MutableMapping
    #
    # def __getitem__(self, item):
    #     return self.__dct__[item]
    #
    # def __setitem__(self, key, value):
    #     self.__dct__[key] = value
    #
    # def __delitem__(self, key):
    #     del self.__dct__[key]
    #
    # def __iter__(self):
    #     return iter(self.__dct__)
    #
    # def __len__(self):
    #     return len(self.__dct__)

    # object base

    def __str__(self):
        return dict.__str__(self)

    def __repr__(self):
        return "ResultsBag:\n" + dict.__repr__(self)

    def __hash__(self):
        return id(self)


# Just to be sure
assert isinstance(ResultsBag(), dict)


def _results_bag_fixture_impl(bag_type=None,                  # type: Type[Any]
                              ):
    """
    Implementation of a results bag fixture. It creates a `ResultsBag` by default, or an object of custom `bag_type`.

    Note: we do not measure time here anymore because it is less precise than pytest duration

    :param bag_type: the type of object to create as results bag. Default: `ResultsBag`
    :return:
    """

    # Create the results bag
    bag_type = ResultsBag if bag_type is None else bag_type
    results_bag = bag_type()

    # Yield it - note: we do not measure time anymore because it is less precise than pytest duration
    # start = datetime.now()
    yield results_bag
    # end = datetime.now()

    # Fill execution time
    # results_bag[exec_time_key] = (end - start).total_seconds()


def create_results_bag_fixture(storage,                        # type: Union[str, Dict[str, Any]]
                               name='results_bag',             # type: str
                               bag_type=None,                  # type: Type[Any]
                               ):
    """
    Creates a "results bag" fixture with name <name> stored in the given storage (under key=<name>). By default results
    bags are instances of `ResultsBag` but you can provide another `bag_type` if needed.

    :param storage: a dict-like object or a fixture name corresponding to a dict-like object. in this dictionary, a new
        entry will be added for the fixture. This entry will contain a dictionary <test_id>: <fixture_value> for each
        test node.
    :param name: the name associated with the stored fixture in the global storage. By default this is 'results_bag'.
    :param bag_type: the type of object to create as a results bag. Default: `ResultsBag`
    :return:
    """
    # Create the same function than _results_bag_fixture_impl but with preset arguments (same than functools.partial)
    def _results_bag():
        gen = _results_bag_fixture_impl(bag_type=bag_type)
        for res in gen:
            yield res

    # Create one result bag per step if needed (if pytest_harvest is present)
    _results_bag = one_per_step(_results_bag)

    # Declare that this fixture should be saved
    _results_bag = saved_fixture(storage, key=name)(_results_bag)

    # Decorate manually as a fixture
    _results_bag.__name__ = name
    results_bag_fixture = yield_fixture(_results_bag)

    return results_bag_fixture



# META
# {'passed': 2, 'failed': 0}
# END META

import os
import pytest

from collections import OrderedDict
from random import random

from pytest_harvest import saved_fixture, get_fixture_value
from pytest_harvest.common import yield_fixture


# init
this_file_name = os.path.split(__file__)[1]
unique_numbers = [random(), random()]


@pytest.fixture(params=unique_numbers)
@saved_fixture('store')
def my_fix(request):
    """Our saved fixture, that will be saved in the store fixture"""
    # convert the parameter to a string so that the fixture is different from the parameter
    return "my_fix #%s" % request.param


def test_foo(my_fix):
    """"""
    print(my_fix)


# -----------------------------------------


@yield_fixture(scope='session', autouse=True)
def store(request):

    # yield the store fixture
    store = OrderedDict()
    yield store

    # check that this util works
    assert get_fixture_value(request, 'store') == store

    # check that the store contains everything
    assert 'my_fix' in store
    assert len(store['my_fix']) == 2
    assert list(store['my_fix'].keys()) == [item.nodeid for item in request.session.items
                                            if this_file_name in item.nodeid]
    assert list(store['my_fix'].values()) == [("my_fix #%s" % n) for n in unique_numbers]



"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""
from six import raise_from
from os import path

from setuptools import setup, find_packages

here = path.abspath(path.dirname(__file__))

# *************** Dependencies *********
INSTALL_REQUIRES = ['decorator', 'funcsigs;python_version<"3.3"']
DEPENDENCY_LINKS = []
SETUP_REQUIRES = ['pytest-runner', 'setuptools_scm', 'pypandoc', 'pandoc']
TESTS_REQUIRE = ['pytest', 'pytest-logging', 'pytest-cov', 'pytest-steps', 'pandas', 'tabulate']
EXTRAS_REQUIRE = {}

# simple check
try:
    from setuptools_scm import get_version
except Exception as e:
    raise_from(Exception('Required packages for setup not found. You may wish you execute '
                         '"pip install -r ci_tools/requirements-setup.txt" to install them or alternatively install '
                         'them manually using conda or other system. The list is : ' + str(SETUP_REQUIRES)), e)

# ************** ID card *****************
DISTNAME = 'pytest-harvest'
DESCRIPTION = 'Store data created during your pytest tests execution, and retrieve it at the end of the session, ' \
              'e.g. for applicative benchmarking purposes.'
MAINTAINER = 'Sylvain MARIE'
MAINTAINER_EMAIL = 'sylvain.marie@schneider-electric.com'
URL = 'https://github.com/smarie/python-pytest-harvest'
LICENSE = 'BSD 3-Clause'
LICENSE_LONG = 'License :: OSI Approved :: BSD License'

version_for_download_url = get_version()
DOWNLOAD_URL = URL + '/tarball/' + version_for_download_url

KEYWORDS = 'pytest test result store fixture collect benchmark artifact session'
# --Get the long description from the README file
# with open(path.join(here, 'README.md'), encoding='utf-8') as f:
#    LONG_DESCRIPTION = f.read()
try:
    import pypandoc
    LONG_DESCRIPTION = pypandoc.convert(path.join(here, 'docs', 'long_description.md'), 'rst').replace('\r', '')
except(ImportError):
    from warnings import warn
    warn('WARNING pypandoc could not be imported - we recommend that you install it in order to package the '
         'documentation correctly')
    LONG_DESCRIPTION = open('README.md').read()

# ************* VERSION A **************
# --Get the Version number from VERSION file, see https://packaging.python.org/single_source_version/ option 4.
# THIS IS DEPRECATED AS WE NOW USE GIT TO MANAGE VERSION
# with open(path.join(here, 'VERSION')) as version_file:
#    VERSION = version_file.read().strip()
# OBSOLETES = []

setup(
    name=DISTNAME,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    # version=VERSION, NOW HANDLED BY GIT

    maintainer=MAINTAINER,
    maintainer_email=MAINTAINER_EMAIL,

    license=LICENSE,
    url=URL,
    download_url=DOWNLOAD_URL,

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 5 - Production/Stable',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Testing',

        # Pick your license as you wish (should match "license" above)
        LICENSE_LONG,

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        # 'Programming Language :: Python :: 2',
        # 'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        # 'Programming Language :: Python :: 3',
        # 'Programming Language :: Python :: 3.3',
        # 'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',

        'Framework :: Pytest'
    ],

    # What does your project relate to?
    keywords=KEYWORDS,

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),

    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=INSTALL_REQUIRES,
    dependency_links=DEPENDENCY_LINKS,

    # we're using git
    use_scm_version=True, # this provides the version + adds the date if local non-commited changes.
    # use_scm_version={'local_scheme':'dirty-tag'}, # this provides the version + adds '+dirty' if local non-commited changes.
    setup_requires=SETUP_REQUIRES,

    # test
    # test_suite='nose.collector',
    tests_require=TESTS_REQUIRE,

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require=EXTRAS_REQUIRE,

    # obsoletes=OBSOLETES

    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    # package_data={
    #     'sample': ['package_data.dat'],
    # },

    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    # data_files=[('my_data', ['data/data_file'])],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    # entry_points={
    #     'console_scripts': [
    #         'sample=sample:main',
    #     ],
    # },

    # the following makes a plugin available to pytest
    entry_points={"pytest11": ["harvest = pytest_harvest.plugin"]},
)



# META
# {'passed': 11, 'skipped': 1, 'failed': 1}
# END META
import os
from itertools import product

import pytest
from pytest_harvest import get_session_synthesis_dct
from pytest_harvest.common import yield_fixture


# init
this_file_name = os.path.split(__file__)[1]

# parametrized fixture
fixture_params = [1, 2]


@pytest.fixture(params=fixture_params)
def a_number_str(request):
    return "my_fix #%s" % request.param


# unparametrized fixture
@pytest.fixture
def dummy():
    return "hey there !"


# parametrized test
test_params = ['hello', 'world']


@pytest.mark.parametrize('p', test_params, ids=str)
def test_foo(p, a_number_str, dummy):
    print(p + a_number_str)


@pytest.mark.parametrize('flatten', [False, True], ids="flatten={}".format)
@pytest.mark.parametrize('durations_in_ms', [False, True], ids="duration_in_ms={}".format)
def test_foo_synthesis_all_options(request, flatten, durations_in_ms):
    """
    Tests that the synthesis is ok

    :param request:
    :return:
    """
    # Get the synthesis dictionary concerning `test_foo`
    synth_dct = get_session_synthesis_dct(request.session, status_details=True, flatten=flatten, filter=test_foo,
                                          durations_in_ms=durations_in_ms)

    # from pprint import pprint
    # pprint(dict(synth_dct))

    durations_unit = ('ms' if durations_in_ms else 's')

    # Check the returned dictionary contents
    prefix = '' if flatten else 'pytest_'
    expected_keys = {'pytest_obj',
                     prefix + 'status',
                     prefix + 'duration_' + durations_unit}
    stages = ['setup', 'call', 'teardown']
    if not flatten:
        expected_keys.update({prefix + 'status_details', prefix + 'params'})
    else:
        expected_keys.update({(prefix + 'status__' + stage) for stage in stages})
        # add parameters
        expected_keys.update({mark.args[0] for mark in test_foo.parametrize})
        # add parametrized fixtures
        expected_keys.update({parametrized_fixture.__name__ for parametrized_fixture in [a_number_str]})

    # compute the parameter values for all tests in order
    params = list(product(fixture_params, test_params))

    for i, (nodeid, nodeinfo) in enumerate(synth_dct.items()):

        # check that all keys are present
        assert set(nodeinfo.keys()) == expected_keys

        # main test information
        assert nodeinfo['pytest_obj'] == test_foo  # check that the filter worked
        assert nodeinfo[prefix + 'status'] == 'passed'
        assert nodeinfo[prefix + 'duration_' + durations_unit] >= 0

        # test status details
        if not flatten:
            assert set(nodeinfo[prefix + 'status_details'].keys()) == set(stages)
        for step in stages:
            if flatten:
                step_info = nodeinfo[prefix + 'status__' + step]
            else:
                step_info = nodeinfo[prefix + 'status_details'][step]
            assert len(step_info) == 2
            assert step_info[0] == 'passed'
            assert step_info[1] >= 0

        # parameter values
        if flatten:
            param_dct = nodeinfo
        else:
            assert set(nodeinfo[prefix + 'params'].keys()) == {'p', 'a_number_str'}
            param_dct = nodeinfo[prefix + 'params']

        assert param_dct['a_number_str'] == params[i][0]
        assert param_dct['p'] == params[i][1]


# skipped test
def test_skipped():
    pytest.skip("normal, intended skip here")


# failing test
def test_failing():
    pytest.fail("normal, intended failure here")


def test_synthesis_skipped(request):
    """ Tests that the synthesis concerning the skipped test is correct """
    synth_dct = get_session_synthesis_dct(request.session, filter=[test_skipped])
    for test_id, v in synth_dct.items():
        assert v['pytest_status'] == 'skipped'


def test_synthesis_failed(request):
    """ Tests that the synthesis concerning the failed test is correct """
    synth_dct = get_session_synthesis_dct(request.session, filter=[test_failing])
    for test_id, v in synth_dct.items():
        assert v['pytest_status'] == 'failed'


def test_synthesis_contains_everything(request):
    """ Tests that the synthesis contains all test nodes """
    # retrieve session results
    synth_dct = get_session_synthesis_dct(request.session, filter_incomplete=False)

    # asserts
    these_tests = [item.nodeid for item in request.session.items if this_file_name in item.nodeid]

    # check that synth_dct contains all these test nodes
    missing = set(these_tests) - set(synth_dct.keys())
    assert len(missing) == 0


@yield_fixture(scope='session', autouse=True)
def make_synthesis(request):
    """This checks that the session-scoped fixture teardown hook works as well"""
    yield
    test_synthesis_contains_everything(request)
